/* wptClipEditDlg.cpp - Clipboard Editor dialog
 *	Copyright (C) 2000-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <stdio.h>

#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptErrors.h"
#include "wptGPG.h"
#include "wptNLS.h"
#include "resource.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptKeylist.h"


/* XXX: it would be really nice to have tempest prevention fonts. */

void verify_get_clip_info (gpgme_signature_t sig, 
			   char **r_header, char **r_footer);


/* Load the clipboard contents into the text field in @dlg.
   Optinally @header and @footer can be used to set a header and footer.
   Return value: 0 on success. */
static int
load_clipboard (HWND dlg, const char *header, const char *footer)
{
    HANDLE clipmem;
    char *cliptext;
    char *p, *pp;

    CloseClipboard (); /* make sure it is closed. */
    if (OpenClipboard (NULL) == FALSE) {
	msg_box (dlg, winpt_strerror (WPTERR_CLIP_OPEN), _("Clipboard"), MB_ERR);
	return -1;
    }
    clipmem = GetClipboardData (CF_TEXT);
    if (clipmem == NULL) {
	CloseClipboard ();
	return -1;
    }
    
    cliptext = (char *) GlobalLock (clipmem);
    if (cliptext == NULL || !strlen (cliptext)) {    
	CloseClipboard ();
	return -1;
    }

    p = new char [strlen (cliptext)+2];
    if (!p)    
	BUG (0);	
    strcpy (p, cliptext);
    
    GlobalUnlock (clipmem);
    CloseClipboard ();

    if (header && footer) {
	const char *fmt = "%s\r\n%s\r\n%s";
	int len;

	len = strlen (header) + strlen (footer) + strlen (fmt) + strlen (p);
	pp = new char[len + 1];
	if (!pp)
	    BUG (0);
	_snprintf (pp, len, fmt, header, p, footer);
	SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, pp);
	free_if_alloc (pp);
    }
    else
	SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p);
    free_if_alloc (p);
    return 0;
}


/* Load the clipboard from the selected file. */
static int
load_clipboard_from_file (HWND dlg)
{
    OPENFILENAME open;
    FILE *fp;
    char *p;
    char file[MAX_PATH+64] = "";
    DWORD size;
    int id;

    memset (&open, 0, sizeof (open));
    open.lStructSize = sizeof (OPENFILENAME);
    open.hInstance = glob_hinst;
    open.lpstrTitle = _("File Open");
    open.lpstrFilter = "All Files (*.*)\0*.*\0\0";
    open.hwndOwner = dlg;	    
    open.lpstrFile = file;		    
    open.nMaxFile = DIM (file)-1;		    
    open.Flags = OFN_FILEMUSTEXIST|OFN_PATHMUSTEXIST;
	    
    if (GetOpenFileName (&open)) {
	if( (size=get_file_size (file)) > MAX_CLIPTEXT_SIZE) {
	    id = msg_box (dlg, _("The file you want to add is very large.\n"
			         "Continue?"), _("Clipboard"), 
			  MB_INFO|MB_YESNO);
	    if (id == IDNO)
		return -1;
	}
	fp = fopen (file, "rb");
	if (!fp) {
	    msg_box (dlg, winpt_strerror (WPTERR_FILE_OPEN), 
		     _("Clipboard"), MB_ERR);
	    return -1;
	}
	p = new char[size+1];
	if (!p)
	    BUG (0);
	memset (p, 0, size+1);
	size = fread (p, 1, size, fp);
	fclose (fp);
	SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p);	
	free_if_alloc (p);
    }
    return 0;
}


/* Save the contents of the clipboard to the selected file. */
static int
save_clipboard_to_file (HWND dlg)
{
    OPENFILENAME save;
    FILE *fp;
    char *p=NULL;
    char file[300] = "";
    DWORD nbytes;

    memset (&save, 0, sizeof (save));	    
    save.lStructSize = sizeof (OPENFILENAME);	    
    save.hInstance = glob_hinst;	    
    save.lpstrTitle = _("File Save");	    
    save.lpstrFile = "All Files (*.*)\0*.*\0\0";
    save.hwndOwner = dlg;
    save.lpstrFile = file;
    save.nMaxFile = DIM (file) - 1;
    save.Flags = OFN_OVERWRITEPROMPT;

    if (GetSaveFileName (&save)) {
	fp = fopen (file, "wb");
	if (!fp) {
	    msg_box (dlg, winpt_strerror (WPTERR_FILE_CREAT), 
		     _("Clipboard"), MB_ERR);
	    return -1;
	}
	nbytes = SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, 
				     WM_GETTEXTLENGTH, 0, 0);
	if (nbytes > 0) {
	    p = new char[nbytes+1];
	    if (!p)
		BUG (NULL);
	    GetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, p, nbytes);
	    fwrite (p, 1, nbytes, fp);
	    
	}	   
	fclose (fp);
	free_if_alloc (p);
	msg_box (dlg, _("Data successfully written to file."),
		 _("Clipboard"), MB_OK);
    }
    return 0;
}


/* Dialog box procedure for the clipboard editor. */
BOOL CALLBACK 
clip_edit_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    HMENU menu;
    char *head=NULL, *foot=NULL;
    gpgme_data_t txt;

    switch (msg)  {
    case WM_INITDIALOG:		
	/* If there is a param, we expect a signature and initialize
	   the header and footer variable accordingly. */
	if (lparam)
	    verify_get_clip_info ((gpgme_signature_t)lparam, &head, &foot);
	load_clipboard (dlg, head, foot);
	free_if_alloc (head);
	free_if_alloc (foot);

	menu = LoadMenu (glob_hinst, (LPCTSTR)IDR_WINPT_CLIPEDIT);
	set_menu_text (menu, ID_CLIPEDIT_SEND, _("&Copy"));
	set_menu_text (menu, ID_CLIPEDIT_CLEAR, _("Clea&r"));
	set_menu_text (menu, ID_CLIPEDIT_QUOTE, _("&Quote")); 
	set_menu_text (menu, ID_CLIPEDIT_OPEN, _("&Open..."));	
	set_menu_text (menu, ID_CLIPEDIT_SAVE, _("&Save..."));
	set_menu_text (menu, ID_CLIPEDIT_PASTE, _("&Paste"));
	set_menu_text (menu, ID_CLIPEDIT_ENC, _("&Encrypt"));
	set_menu_text (menu, ID_CLIPEDIT_DEC, _("&Decrypt"));
	SetMenu (dlg, menu);
	SetWindowText (dlg, _("Clipboard Editor"));
	SetDlgItemText (dlg, IDC_CLIPEDIT_SEND, _("&Copy"));
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	return TRUE;
        
    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    return TRUE;

	case IDOK:
	case ID_CLIPEDIT_QUIT:
	    EndDialog (dlg, TRUE);
	    return TRUE;
            
	case ID_CLIPEDIT_QUOTE:
	    SendMessage (dlg, WM_COMMAND, ID_CLIPEDIT_COPY, 0);
	    if (gpg_data_new_from_clipboard (&txt, 0))
		return TRUE;
	    gpg_data_mail_quote (&txt);
	    gpg_data_release_and_set_clipboard (txt, 0);
	    load_clipboard (dlg, head, foot);
	    return TRUE;

	case ID_CLIPEDIT_COPY:
	case IDC_CLIPEDIT_SEND:
	    SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, EM_SETSEL, 0, -1);
	    SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, WM_COPY, 0, 0);
	    SendDlgItemMessage (dlg, IDC_CLIPEDIT_CLIPTEXT2, EM_SETSEL, -1, 0);
	    return TRUE;
            
	case ID_CLIPEDIT_CLEAR:
	    if (OpenClipboard (NULL) == FALSE) {
		msg_box (dlg,  winpt_strerror (WPTERR_CLIP_OPEN), 
			 _("Clipboard"), MB_ERR);
		return TRUE;
	    }			
	    if (EmptyClipboard () == FALSE)		
		msg_box (dlg, winpt_strerror (WPTERR_CLIP_EMPTY), 
			 _("Clipboard"), MB_ERR);
	    else
		SetDlgItemText (dlg, IDC_CLIPEDIT_CLIPTEXT2, "");
	    CloseClipboard ();
	    return TRUE;

	case ID_CLIPEDIT_PASTE:
	    load_clipboard (dlg, NULL, NULL);
	    break;
            
	case ID_CLIPEDIT_OPEN:
	    if (!load_clipboard_from_file (dlg))
		PostMessage (dlg, WM_COMMAND, MAKEWPARAM (IDC_CLIPEDIT_SEND, 0), 0);
	    return TRUE;
            
	case ID_CLIPEDIT_SAVE:
	    save_clipboard_to_file (dlg);	    
	    return TRUE;

	case ID_CLIPEDIT_ENC:
	    SendMessage (dlg, WM_COMMAND, ID_CLIPEDIT_COPY, 0);
	    SendMessage (glob_hwnd, WM_COMMAND, ID_WINPT_ENCRYPT, 0);
	    load_clipboard (dlg, NULL, NULL);
	    SetForegroundWindow (dlg);
	    break;

	case ID_CLIPEDIT_DEC:
	    SendMessage (glob_hwnd, WM_COMMAND, ID_WINPT_DECRYPT_VERIFY, 0);
	    load_clipboard (dlg, NULL, NULL);
	    SetForegroundWindow (dlg);
	    break;
	}
	break;
    }
    
    return FALSE;
}
