/* wptCurrWnd.cpp - Current window code
 *	Copyright (C) 2001-2004, 2006, 2007 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>

#include "wptW32API.h"
#include "wptErrors.h"
#include "wptVersion.h"

#define KEYDOWN(vcode, scode) keybd_event ((vcode), (scode), 0, 0)
#define KEYUP(vcode, scode)   keybd_event ((vcode), (scode), KEYEVENTF_KEYUP, 0)


/* PTD prototypes from the DLL. */
extern "C" HWND PTD_get_curr_hwnd (void);
extern "C" BOOL PTD_keyb_send (UINT *keys, UINT n_keys);
extern "C" int PTD_is_used (void);


/* Clear the clipboard contents. */
static void
clip_clear (void)
{
    OpenClipboard (NULL);
    EmptyClipboard ();
    CloseClipboard ();
}


/* Check if the clipboard returns any text.
   Return value: 0 on success. */
static int
clip_check (void)
{
    HANDLE clipmem;

    /* Little backoff time to make sure the clipboard now really 
       contains all data. */
    Sleep (400);
		
    if (OpenClipboard (NULL) == FALSE) 
	return WPTERR_CLIP_OPEN;
    clipmem = GetClipboardData (CF_TEXT);
    if (clipmem == NULL) {
	CloseClipboard ();
	return WPTERR_CLIP_GET;	
    }
    CloseClipboard ();
    return 0;
}


/* Fabricate a "select all" keystroke sequence. */
static void
window_msg_selectall (void)
{
    unsigned int keys[4];

    if (PTD_is_used ()) {
	keys[0] = VK_CONTROL | 0x8000;
	keys[1] = 'A' | 0x8000;	
	keys[2] = 'A';	
	keys[3] = VK_CONTROL;	
	PTD_keyb_send (keys, 4);

    }
    else {
	KEYDOWN (VK_CONTROL, 0x1d);
	KEYDOWN ((BYTE)'A', 0x1e);
	KEYUP ((BYTE)'A', 0x1e);
	KEYUP (VK_CONTROL, 0x1d);
    }
}


/* Fabricate a "paste" keystrok sequence. */
static void
window_msg_paste (void)
{
    unsigned int keys[4];

    if (PTD_is_used ()) {
	keys[0] = VK_CONTROL | 0x8000;
	keys[1] = 'V' | 0x8000;
	keys[2] = 'V';
	keys[3] = VK_CONTROL;
	PTD_keyb_send (keys, 4);
    }
    else {
	KEYDOWN (VK_CONTROL, 0x1d);
	KEYDOWN ((BYTE)'V', 0x2f);
	KEYUP ((BYTE)'V', 0x2f);
	KEYUP (VK_CONTROL, 0x1d);
    }
}


/* Fabricate a "copy" keystroke sequence. */
static void
window_msg_copy (void)
{
    unsigned int keys[4];

    if (PTD_is_used ()) {
	keys[0] = VK_CONTROL | 0x8000;
	keys[1] = 'C' | 0x8000;
	keys[2] = 'C';
	keys[3] = VK_CONTROL;
	PTD_keyb_send (keys, 4);
    }
    else {
	KEYDOWN (VK_CONTROL, 0x1d);
	KEYDOWN ((BYTE)'C', 0x2e);
	KEYUP ((BYTE)'C', 0x2e);
	KEYUP (VK_CONTROL, 0x1d);
    }
}


/* Set the position inside the edit control @hwnd. */
static inline void
window_msg_em_set_pos (HWND hwnd, int begin)
{
    SendMessage (hwnd, EM_SETSEL, 0, begin? 0 : -1);
}


/* Get the current window and return it in @r_main.
   Return value: The Window with the keyboard focus. */
static HWND
get_current_wnd (HWND main, int use_hotkey, HWND *r_main)
{
    HWND prev, fg;

    if (use_hotkey)
	prev = GetForegroundWindow ();
    else {
	*r_main = GetNextWindow (main, GW_HWNDNEXT);
	return PTD_get_curr_hwnd ();
    }

    SetForegroundWindow (prev);
    AttachThreadInput (GetCurrentThreadId(),
		       GetWindowThreadProcessId (prev, NULL),
		       TRUE);
    fg = GetFocus ();
    AttachThreadInput (GetCurrentThreadId (),
			GetWindowThreadProcessId (prev, NULL),
			FALSE);

    *r_main = prev;
    return fg;
}


/* Retrieve the contents of the current window based on
   the old window handle @old_hwnd and store the information
   in @ctx. 
   A return value of 0 indicates a problem. */
int
copy_window_content (HWND old_hwnd, HWND *r_main, HWND *r_focus,
		     int use_hotkey)
{
    HWND hwnd;
    int rc;
	
    rc = WPTERR_GENERAL;
    hwnd = get_current_wnd (old_hwnd, use_hotkey, r_main);
    if (!hwnd)
	return rc;
    *r_focus = hwnd;
    clip_clear ();
	
    AttachThreadInput (GetCurrentThreadId (),
		       GetWindowThreadProcessId (hwnd, NULL),
		       TRUE);
    SetFocus (hwnd);
    /* First we try to send a simple clipboard copying command and check
       if the clipboard contains some text. */
    SendMessage (hwnd, WM_COPY, 0, 0);
    if (!clip_check ()) {
	rc = 0; 
	goto leave;
    }
    /* Then the check if the window is an edit control */
    window_msg_em_set_pos (hwnd, 0);
    SendMessage (hwnd, WM_COPY, 0, 0);
    if (!clip_check ()) {
	rc = 0; 
	goto leave;
    }
	
    /* The last try is to send a mark all and copy to clipboard command */
    window_msg_selectall ();
    window_msg_copy ();
    if (!clip_check ())
	rc = 0;

leave:
    AttachThreadInput (GetCurrentThreadId (),
		       GetWindowThreadProcessId (hwnd, NULL),
		       FALSE);
    return rc;
}


/* Restore the clipboard contents to the window handle given
   in @ctx and reset the current window to @old_hwnd if possible. */
int
paste_window_content (HWND old_hwnd, HWND main, HWND focus)
{
    HWND hwnd, lost;
	
    hwnd = focus;
    
    /* Restore window when in minimized state. */
    if (IsIconic (main))
	ShowWindow (hwnd, SW_SHOWNORMAL);
    SetForegroundWindow (hwnd);
	
    AttachThreadInput (GetCurrentThreadId (),
		       GetWindowThreadProcessId (hwnd, NULL),
		       TRUE);
	
    lost = SetFocus (hwnd);

    window_msg_em_set_pos (hwnd, 0);
    window_msg_paste ();
    window_msg_em_set_pos (hwnd, 1);
	
    AttachThreadInput (GetCurrentThreadId(),
		       GetWindowThreadProcessId (hwnd, NULL),
		       FALSE);
	
    SetForegroundWindow (lost);
    return 0;
}
