/* wptKeysignDlg.cpp - Key signing dialog
 *	Copyright (C) 2001-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>
#include <time.h>
#include <assert.h>

#include "resource.h"
#include "wptGPG.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptTypes.h"
#include "wptErrors.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptUTF8.h"
#include "wptRegistry.h"
#include "wptKeylist.h"
#include "wptKeyEdit.h"
#include "StringBuffer.h"


void key_unload_photo (HWND dlg);
int  key_load_photo (HWND dlg, gpgme_key_t key, gpgme_validity_t *r_valid);
const char* get_photo_tmpname (HWND dlg);

/* Return a beautified printable fingerprint of @fpr. */
static const char*
get_printable_fpr (const char *fpr)
{
    static char pfpr[64];
    int pos = 0;
    size_t i;

    for (i = 0; i < strlen (fpr); i += 4) {
	pfpr[pos++] = fpr[i];
	pfpr[pos++] = fpr[i+1];
	pfpr[pos++] = fpr[i+2];
	pfpr[pos++] = fpr[i+3];
	pfpr[pos++] = ' ';
    }
    return pfpr;
}


/* Return human friendly information about the key @key. */
const char*
get_keyinfo (gpgme_key_t key)
{
    static char buf[64+16];
    struct winpt_key_s k;

    memset (&k, 0, sizeof (k));
    if (winpt_get_seckey (key->subkeys->keyid, &k))
	BUG (0);
    _snprintf (buf, DIM (buf)-1-16, _("%d-bit %s key, ID 0x%s"),
	       key->subkeys->length,
	       get_key_pubalgo (key->subkeys->pubkey_algo),
	       key->subkeys->keyid+8);
    if (k.ext->gloflags.divert_to_card)
    	strcat (buf, " (Card)");
    return buf;
}


/* Fill the secret key combo-box with all entries from the cache.
   @dlg is the handle to the combo-box. @keyid show which key to skip.
   Return value: 0 on success. */
static int
do_fill_seckeylist (HWND dlg, const char *keyid)
{
    gpg_keycache_t sec;
    gpgme_key_t pk, defkey;
    const char *s;
    char *uid;
    int i = 0, n = 0, curr_sel = 0;

    sec = keycache_get_ctx (0);
    gpg_keycache_get_default_key (sec, &defkey);
    gpg_keycache_rewind (sec);
    while (!gpg_keycache_next_key (sec, 1, &pk)) {
	StringBuffer p;

	if (!pk)
	    continue;
	s = pk->subkeys->keyid;
	if (!strcmp (s, keyid))
	    continue;
	/* skip all ElGamal sign+encrypt keys */
	if (pk->subkeys->pubkey_algo == GPGME_PK_ELG)
	    continue;
	/* make sure the public key is okay not: revoked, expired or disabled. */
	if (pk->expired ||pk->revoked || pk->disabled)
	    continue;
	s = pk->uids->name;
	if (!s)
	    continue;
	if (defkey && !strcmp (defkey->subkeys->keyid, pk->subkeys->keyid))
	    curr_sel = i;

	uid = utf8_to_native (s);
	p = uid;
	p = p + " (" + get_keyinfo (pk) + ")";
	SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, 
			    CB_ADDSTRING, i, (LPARAM)(char *)p.getBuffer ());
	SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, 
			    CB_SETITEMDATA, i++, (LPARAM)(DWORD)pk);
	safe_free (uid);
	n++;
    }
    SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, 
			CB_SETCURSEL, (WPARAM)curr_sel, 0);
    if (!n)
	return -1;
    return 0;
}


/* Check if the selected key is protected and en- or disable the 
   passphrase control. */
static void
do_check_protection (HWND dlg)
{
    gpgme_key_t key;
    struct winpt_key_s k;
    int idx, protec;    

    idx = SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, CB_GETCURSEL, 0, 0);
    key = (gpgme_key_t)SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, 
					   CB_GETITEMDATA, (WPARAM)idx, 0);
    if (key) {
	if (winpt_get_seckey (key->subkeys->keyid, &k))
	    BUG (0);
	protec = k.is_protected;
	if (!protec)
	    protec = k.ext->gloflags.divert_to_card;
	EnableWindow (GetDlgItem (dlg, IDC_KEYSIGN_PASSPHRASE), protec? TRUE : FALSE);
    }
}


/* Show the photographic ID of the key in a separate dialog window. */
BOOL CALLBACK
showphoto_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static winpt_key_t key;
    char buf[200];

    switch (msg) {
    case WM_INITDIALOG:
	key = (winpt_key_t)lparam;
	if (!key)
	    BUG (0);
	key_load_photo (dlg, key->ctx, NULL);
	_snprintf (buf, DIM (buf)-1, _("Photo of %s"), key->uid);
	SetWindowText (dlg, buf);
	SetForegroundWindow (dlg);
	break;

    case WM_PAINT:
	if (key->ext && key->ext->attrib.len > 0) {
	    POINT p;
	    p.x = p.y = 0;
	    jpg_show (GetDlgItem (dlg, IDC_SHOWPHOTO_IMG), 
			    &p, get_photo_tmpname (dlg));
	}
	break;

    case WM_DESTROY:
	key_unload_photo (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDCANCEL:
	case IDOK:
	    EndDialog (dlg, TRUE);
	    break;
	}
	break;
    }

    return FALSE;
}


/* Dialog box procedure to choose the signature class. */
BOOL CALLBACK
sig_class_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    HWND parent;
    int sig_class = 0;

    switch (msg) {
    case WM_INITDIALOG:
	parent = (HWND)lparam;
	SetWindowText (dlg, _("Choose Signature Class"));
	SetDlgItemText (dlg, IDC_SIGCLASS_TITLEINF, 
			_("How carefully have you verified the key you are about to sign actually belongs to the person? If you don't know what to anwser, use \"0\"."));
	SetDlgItemText (dlg, IDC_SIGCLASS_CLASS0, 
			_("(0) I will not answer (default)"));
	SetDlgItemText (dlg, IDC_SIGCLASS_CLASS1, 
			_("(1) I have not checked at all."));
	SetDlgItemText (dlg, IDC_SIGCLASS_CLASS2, 
			_("(2) I have done causal checking."));
	SetDlgItemText (dlg, IDC_SIGCLASS_CLASS3, 
			_("(3) I have done very careful checking."));
	CheckDlgButton (dlg, IDC_SIGCLASS_CLASS0, BST_CHECKED);
	SetForegroundWindow (dlg);
	center_window (dlg, parent? parent : NULL);
	return TRUE;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    if (IsDlgButtonChecked (dlg, IDC_SIGCLASS_CLASS0))
		sig_class = 0;
	    else if (IsDlgButtonChecked (dlg, IDC_SIGCLASS_CLASS1))
		sig_class = 1;
	    else if (IsDlgButtonChecked (dlg, IDC_SIGCLASS_CLASS2))
		sig_class = 2;
	    else if (IsDlgButtonChecked (dlg, IDC_SIGCLASS_CLASS3))
		sig_class = 3;
	    else
		sig_class = 0;
	    EndDialog (dlg, sig_class);
	    return TRUE;

	case IDCANCEL:
	    EndDialog (dlg, 0);
	    return TRUE;
	}
	break;
    }

    return FALSE;
}


/* Return the humand friendly expiration date of @key. */
static const char*
get_expire_date (gpgme_key_t key)
{
    u32 u = key->subkeys->expires;
    if (!u)
	return _("never");
    return get_key_expire_date (u);
}


/* Check if the given system time @st points to today. */
static int
date_is_today (SYSTEMTIME *st)
{
    time_t t;
    struct tm *tm;

    t = time (NULL);
    tm = localtime (&t);
    if (st->wDay == tm->tm_mday &&
	st->wYear == tm->tm_year+1900 &&
	st->wMonth == tm->tm_mon+1)
	return -1;
    return 0;
}


/* Display all needed information about the key @key in the dialog @dlg. */
static void
display_keyinfo (HWND dlg, winpt_key_t key)
{
    StringBuffer kinf;
    struct native_uid_s *u;
    char tmp[256];
    int len=0;
   
    _snprintf (tmp, DIM (tmp) -1,
                   _("pub %d/0x%s   created: %s    expires: %s\n\n"
                     "Primary key fingerprint: %s\n\n"),
                   key->ctx->subkeys->length,
		   key->ctx->subkeys->keyid+8,
                   get_key_created (key->ctx->subkeys->timestamp),
		   get_expire_date (key->ctx),
                   get_printable_fpr (key->ctx->subkeys->fpr));     
    kinf = tmp;    
    for (len=0, u = key->ext->uids; u; u = u->next) {
	if (u->revoked)
	    continue;
	if (len++ > 2)
	    break;
	kinf = kinf + "\t\"";
	if (strlen (u->uid) > 128) {
	    kinf.add(u->uid, 128);
	    kinf += "...";
	}
	else
	    kinf += u->uid;
	kinf += "\"\n";
    }
    kinf = kinf + _("\n\nAre you really sure that you want to sign this key with YOUR key?\n");
    SetDlgItemText (dlg, IDC_KEYSIGN_INFOS, kinf.getBuffer ());
}


/* Perform the actual procedure to sign the key. */
static BOOL
on_click_ok (HWND dlg, winpt_key_t key)
{
    gpgme_key_t k;
    gpgme_error_t err;
    SYSTEMTIME st;
    GpgKeyEdit ke;
    int sig_class = 0;
    char keymsg[64], *pwd;
    int type, expires=0, idx;

    if (IsDlgButtonChecked (dlg, IDC_KEYSIGN_ASKLEVEL))
	sig_class = dialog_box_param (glob_hinst,
					(LPCSTR)IDD_WINPT_SIGCLASS, dlg,
					      sig_class_dlg_proc, (LPARAM)dlg,
					      _("Choose Signature Class"), 
					      IDS_WINPT_SIGCLASS);
    type = IsDlgButtonChecked (dlg, IDC_KEYSIGN_LOCAL);
    if (type)
	type = GPG_EDITKEY_LSIGN;
    else
	type = GPG_EDITKEY_SIGN;

	    
    if (reg_prefs.expert && IsDlgButtonChecked (dlg, IDC_KEYSIGN_NREV)) {
	type = GPG_EDITKEY_NRSIGN;
	if (type == GPG_EDITKEY_LSIGN)
	    type = GPG_EDITKEY_NRLSIGN;
    }

    if (IsDlgButtonChecked (dlg, IDC_KEYSIGN_EXPSIG)) {
	DateTime_GetSystemtime (GetDlgItem (dlg, IDC_KEYSIGN_EXPIRES), &st);
	if (date_is_today (&st)) {
	    msg_box (dlg, _("You cannot select today as the expiration date."),
		     _("Key Signing"), MB_INFO);
	    return TRUE;
	}
	else
	    expires = 1;
	_snprintf (keymsg, DIM (keymsg)-1, "%04d-%02d-%02d", 
		    st.wYear, st.wMonth, st.wDay);
    }

    /* XXX: check for --ask-cert-level and --ask-cert-expire in the gpg.conf
	    if an advanced button is checked and offer to add it to the config
	    file. */    
    GetDlgItemText_utf8 (dlg, IDC_KEYSIGN_PASSPHRASE, &pwd);
    
    assert (key->ctx->subkeys->keyid != NULL); /* should be always true. */
    ke.setKeyID (key->ctx->subkeys->keyid);
    ke.setPassphrase (pwd);
    idx = SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, CB_GETCURSEL, 0, 0);
    k = (gpgme_key_t)SendDlgItemMessage (dlg, IDC_KEYSIGN_KEYLIST, 
						 CB_GETITEMDATA, (WPARAM)idx, 0);
    if (k != NULL)
	ke.setLocalUser (k);

    err = ke.signKey (type, sig_class, expires? keymsg : "0");
    sfree_if_alloc (pwd);
    if (err) {
	msg_box (dlg, gpgme_strerror (err), _("Key Signing"), MB_ERR);
	return TRUE;	
    }
    else if (ke.getResult () != 0)
	msg_box (dlg, _("This key is already signed by your key"),
		    _("Key Signing"), MB_INFO);
    else {
	status_box (dlg, _("Key successfully signed."), _("Key Signing"));
	key->update = 1;	
    }
    EndDialog (dlg, TRUE);
    return TRUE;
}


/* Dialog box procedure to sign a key. */
BOOL CALLBACK
keysign_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static winpt_key_t key;
    const char *s;
    
    switch (msg)  {
    case WM_INITDIALOG:
        key = (winpt_key_t) lparam;
	if (!key)
	    BUG (0);        
	s = key->ctx->subkeys->keyid;
	if (do_fill_seckeylist (dlg, s)) {
	    msg_box (dlg, _("No valid secret key found."), _("Key Signing"), MB_ERR);
	    EndDialog (dlg, FALSE);
	    return TRUE;
	}
	do_check_protection (dlg);
	display_keyinfo (dlg, key);
        SetDlgItemText (dlg, IDC_KEYSIGN_LOCAL, _("Sign local only (non exportable signature)"));
	SetDlgItemText (dlg, IDC_KEYSIGN_EXPSIG, _("Signature expires on"));
	SetDlgItemText (dlg, IDC_KEYSIGN_NREV, _("Sign non-revocably"));
	SetDlgItemText (dlg, IDC_KEYSIGN_ASKLEVEL, _("&Ask for certification level"));
	SetDlgItemText (dlg, IDC_KEYSIGN_PWDINF, _("Passphrase"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetDlgItemText (dlg, IDC_KEYSIGN_SHOWIMG, _("&Show photo"));
	SetDlgItemText (dlg, IDC_KEYSIGN_HIDE, _("&Hide Typing"));
	SetWindowText (dlg, _("Key Signing"));
        CheckDlgButton (dlg, IDC_KEYSIGN_LOCAL, BST_CHECKED);
	CheckDlgButton (dlg, IDC_KEYSIGN_EXPSIG, BST_UNCHECKED);
	CheckDlgButton (dlg, IDC_KEYSIGN_ASKLEVEL, BST_UNCHECKED);
	EnableWindow (GetDlgItem (dlg, IDC_KEYSIGN_EXPIRES), FALSE);
	if (reg_prefs.expert == 0)
	    ShowWindow (GetDlgItem (dlg, IDC_KEYSIGN_NREV), SW_HIDE);
	if (key->ext && key->ext->attrib.len == 0)
	    ShowWindow (GetDlgItem (dlg, IDC_KEYSIGN_SHOWIMG), SW_HIDE);
	if (!reg_prefs.gpg.ask_cert_level)
	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIGN_ASKLEVEL), FALSE);
	if (!reg_prefs.gpg.ask_cert_expire)
	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIGN_EXPSIG), FALSE);
	CheckDlgButton (dlg, IDC_KEYSIGN_HIDE, BST_CHECKED);
        SetForegroundWindow (dlg);
	SetFocus (GetDlgItem (dlg, IDC_KEYSIGN_PASSPHRASE));
        return FALSE;
        
    case WM_COMMAND:
	if (HIWORD (wparam) == CBN_SELCHANGE) {
	    do_check_protection (dlg);
	    break;
	}
	if (HIWORD (wparam) == BN_CLICKED && 
	    LOWORD (wparam) == IDC_KEYSIGN_EXPSIG) {
	    int enable = IsDlgButtonChecked (dlg, IDC_KEYSIGN_EXPSIG);

	    EnableWindow (GetDlgItem (dlg, IDC_KEYSIGN_EXPIRES), enable? TRUE : FALSE);
	}
	if (HIWORD (wparam) == BN_CLICKED && 
	    LOWORD (wparam) == IDC_KEYSIGN_HIDE) {
	    HWND hwnd = GetDlgItem (dlg, IDC_KEYSIGN_PASSPHRASE);
	    int hide = IsDlgButtonChecked (dlg, IDC_KEYSIGN_HIDE);

	    SendMessage (hwnd, EM_SETPASSWORDCHAR, hide? '*' : 0, 0);
	    SetFocus (hwnd);
	}

        switch (LOWORD (wparam)) {
        case IDOK:
	    return on_click_ok (dlg, key);
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return TRUE;

	case IDC_KEYSIGN_SHOWIMG:
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SHOWPHOTO,
			    dlg, showphoto_dlg_proc, (LPARAM)key);
	    return TRUE;
        }
        break;
    }
    
    return FALSE;
}
