/* wptMainProc.cpp - Main window procedure
 *	Copyright (C) 2000-2008 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/* Only define the constant if needed. */
#ifndef _WIN32_IE
#define _WIN32_IE 0x0600
#endif

#include <windows.h>
#include <commctrl.h>
#include <io.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptErrors.h"
#include "wptNLS.h"
#include "wptHotkey.h"
#include "wptRegistry.h"
#include "wptKeyserver.h"
#include "wptAgent.h"
#include "wptUTF8.h"
#include "wptKeyManager.h"
#include "wptCard.h"
#include "wptCardEdit.h"


int update_keycache (HWND hwnd);

/* Command for the Key Manager dialog. (0=no command) */
static int cmd = 0;

/* 1 indicates that the clipboard will be emptied o exit. */
static int wipe_contents = 0;


/* Dialog box procedure to display all insecure all ElGamal keys. */
BOOL CALLBACK
elgamal_warn_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    gpg_keycache_t pc;
    gpgme_key_t key;
    char *uid;
    char tmp[128+64+1];

    switch (msg) {
    case WM_INITDIALOG:
	pc = keycache_get_ctx (1);
	while (!gpg_keycache_next_key (pc, 0, &key)) {
	    if (key->subkeys->pubkey_algo == GPGME_PK_ELG) {
		uid = utf8_to_native (key->uids->uid);
		_snprintf (tmp, sizeof (tmp)-1, "(0x%s) %s",
			   key->subkeys->keyid+8, uid);
		SendDlgItemMessage (dlg, IDC_ELGWARN_LIST, LB_ADDSTRING,
				    0, (LPARAM)(const char *)tmp);
		free_if_alloc (uid);
	    }
	}
	gpg_keycache_rewind (pc);
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    EndDialog (dlg, TRUE);
	    break;
	}
	break;
    }
    return FALSE;
}


/* Dialog box procedure to confirm to delete the clipboard contents. */
static BOOL CALLBACK
confirm_delclipboard_dlg (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_INITDIALOG:
	SetWindowText (dlg, _("Delete Clipboard Contents"));
	SetDlgItemText (dlg, IDC_CONFDELCLIP_BRAIN, 
			_("&Remember the answer"));
	SetDlgItemText (dlg, IDC_CONFDELCLIP_INFO, 
			_("Do you want to delete the contents from the clipboard?"));
	SetDlgItemText (dlg, IDYES, _("&Yes"));
	SetDlgItemText (dlg, IDNO, _("&No"));
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDYES:
	    wipe_contents = 1;
	    if (IsDlgButtonChecked (dlg, IDC_CONFDELCLIP_BRAIN))
		set_reg_winpt_flag ("WipeClipboard", 1);
	    EndDialog (dlg, 1);
	    break;

	case IDNO:
	    wipe_contents = 0;
	    if (IsDlgButtonChecked (dlg, IDC_CONFDELCLIP_BRAIN))
		set_reg_winpt_flag ("WipeClipboard", 0);
	    EndDialog (dlg, 0);
	    break;
	}
	break;
    }
    return FALSE;
}


/* Handle the various command requests for the clipboard
   and the current window. */
static void
clip_gpg_dlg (HWND hwnd, UINT id, int *r_upd_window)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    int type;
    int disable_viewer = 0;
    
    if (r_upd_window) {
	*r_upd_window = 1;
	/* for the current window mode it makes no sense to
	   use the clipboard viewer for decryption because the
	   text will be pasted directly into the window. */
	disable_viewer = 1;
    }
    if ((id == ID_WINPT_SIGN || id == ID_WINPT_SIGNENCRYPT)
	&& gnupg_access_keyring (0)) {
	msg_box (hwnd, _("Could not access secret keyring."), _("Sign"), MB_ERR);
	return;
    }
    
    switch (id) {
    case ID_WINPT_SYMENC:
    case ID_WINPT_CURRWND_SYMENC:
        err = gpg_clip_sym_encrypt ();
	if (err && r_upd_window)
	    *r_upd_window = 0;
        break;
        
    case ID_WINPT_ENCRYPT:
    case ID_WINPT_CURRWND_ENCRYPT:
        dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_ENCRYPT, hwnd, 
                          clip_encrypt_dlg_proc, 0,
                          _("Encryption"), IDS_WINPT_ENCRYPT);
        break;
        
    case ID_WINPT_SIGN:
    case ID_WINPT_CURRWND_SIGN:
        dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_SIGN, hwnd, 
                          clip_sign_dlg_proc, 0, 
                          _("Signing"), IDS_WINPT_SIGN);
        break;
        
    case ID_WINPT_SIGNENCRYPT:
    case ID_WINPT_CURRWND_SIGNENCRYPT:
        dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_SIGNENC, hwnd, 
                          clip_signenc_dlg_proc, 0, 
                          _("Sign & Encrypt"), IDS_WINPT_SIGNENC);
        break;
        
    case ID_WINPT_DECRYPT_VERIFY:
    case ID_WINPT_CURRWND_DECRYPT_VERIFY:
	err = gpg_clip_get_pgptype (&type);
	if (err) {
	    msg_box (hwnd, gpgme_strerror (err), _("Clipboard"), MB_ERR);
	    if (r_upd_window)
		*r_upd_window = 0;
	    break;
	}
        if ((type & PGP_MESSAGE) && !(type & PGP_CLEARSIG))
            err = clip_decrypt_dlg (hwnd, disable_viewer? 
					  0 : reg_prefs.use_viewer);
        else if ((type & PGP_SIG) && !(type & PGP_CLEARSIG)) {
            text_input_s input;
            gpgme_data_t sig;
            
            memset (&input, 0, sizeof (input));
            err = gpg_data_new_from_clipboard (&sig, 0);
	    if (err) {
		msg_box (hwnd, gpgme_strerror (err),_("Verify"), MB_ERR);
		if (r_upd_window)
		    *r_upd_window = 0;
		return;
	    }
            input.type = 0;
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_TEXT, hwnd,
                              text_input_dlg_proc, (LPARAM)&input, 
                              _("Text Input"), IDS_WINPT_TEXT);
	    
	    /* the dialog might have changed the clipboard, so we restore
	       the original state (again). */
            gpg_data_release_and_set_clipboard (sig, 0);
            if (input.length > 0) {
                dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_VERIFY, hwnd,
                                  clip_verify_dlg_proc, (LPARAM)&input,
                                  _("Verify"), IDS_WINPT_VERIFY);
                input.length = 0;
		free_if_alloc (input.data);
            }
	    if (r_upd_window)
		*r_upd_window = 0;
        }
        else if (type & PGP_CLEARSIG) {
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_VERIFY, hwnd,
                              clip_verify_dlg_proc, 0,
                              _("Verify"), IDS_WINPT_VERIFY);
	    if (r_upd_window)
		*r_upd_window = 0;
        }
        else if ((type & PGP_PUBKEY) || (type & PGP_SECKEY)) {
	    km_clip_import (GetDesktopWindow (), NULL, NULL);
	    if (r_upd_window)
		*r_upd_window = 0;
	}
        else {
            msg_box (hwnd, _("Unknown OpenPGP type."), _("Clipboard"), MB_ERR);
	    if (r_upd_window)
		*r_upd_window = 0;
	}
    }
}


/* Load the Card Manager with the current card. */
static void
load_card_manager (void)
{
    gpg_card_t card;

    card = gpg_card_load ();
    if (!card)
	return;

    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_CARD_EDIT,
		      GetDesktopWindow (), card_edit_dlg_proc,
		      (LPARAM)card,
		       _("Card Edit"), IDS_WINPT_CARD_EDIT);
    gpg_card_release (card);
}


/* Add the winpt symbol to the task bar. */
static void
add_taskbar_icon (HWND hwnd, NOTIFYICONDATA *nid)
{
    nid->cbSize = sizeof (NOTIFYICONDATA);
    nid->uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    nid->uCallbackMessage = WM_USER;
    nid->hWnd = hwnd;
    nid->hIcon = LoadIcon (glob_hinst, MAKEINTRESOURCE (IDI_WINPT));
    strcpy (nid->szTip, "Windows Privacy Tray v"PGM_VERSION);
    Shell_NotifyIcon (NIM_ADD, nid);
    DestroyIcon (nid->hIcon);
}


/* Initialize all needed common controls. */
static void
init_common_controls (void)
{
    INITCOMMONCONTROLSEX cce;

    /* date control, toolbar, trackbar, ... */
    cce.dwSize = sizeof (INITCOMMONCONTROLSEX);
    cce.dwICC = ICC_DATE_CLASSES|ICC_BAR_CLASSES;
    InitCommonControlsEx (&cce);
}


/* Wipe the clipboard if the user requested it. */
static void
wipe_clipboard (int manual_shutdown)
{
    int has_data = 0;

    if (!gpg_clip_istext_avail (&has_data) && has_data) {
	int chk = get_reg_winpt_flag ("WipeClipboard");
	if (chk == -1 && manual_shutdown)
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_CONFDELCLIP,
			    GetDesktopWindow (), confirm_delclipboard_dlg, 0);
	if (wipe_contents || chk)
	    set_clip_text (NULL, " ", 1);    
	wipe_contents = 0;
    }
}



/* Main message loop for the tray window. */
LRESULT CALLBACK 
winpt_main_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{		
    static NOTIFYICONDATA NID;
    static DWORD cookie;
    static int use_hotkey = 0;
    static int manual_shutdown = 0;
    HWND wnd, cw_main, cw_focus;
    gpgme_error_t err;
    const char *s;
    int rc, set_wc = 0;
    int has_data = 0;

    switch (msg) {
    case WM_CREATE:
        add_taskbar_icon (hwnd, &NID);
        rc = wsock_init ();
        if (rc)
            msg_box (NULL, winpt_strerror (rc), "Winsock2 DLL", MB_ERR);
        if (!reg_prefs.no_hotkeys) {
            rc = hotkeys_register (hwnd);
            if (rc)
                msg_box (hwnd, hotkeys_strerror (), winpt_strerror (rc), MB_ERR);
        }
        rc = PTD_initialize ();
        if (!rc)
            msg_box (hwnd, _("Could not set current window mode hooks."), 
		     _("WinPT Error"), MB_OK);
	mapi_init ();
	init_common_controls ();
	html_help_init (&cookie);
	LoadLibrary ("RichEd32.Dll");
	manual_shutdown = 0;
        break;
        
    case WM_ENDSESSION:
    case WM_DESTROY:
    case WM_CLOSE:
    case WM_QUIT:
	mapi_deinit ();
        wsock_end ();        
        keycache_release (1);
	/* FIXME: There seems to be a problem with the backup at the
	 * end on EFS file systems. I currently have no access to
	 * any EFS partition so I need to do some general investigations
	 * to find out what the problem is.
	 */
	gnupg_backup_keyrings (reg_prefs.auto_backup, 
			       reg_prefs.backup.mode,
			       reg_prefs.backup.include_secr);
	free_reg_prefs ();	
	free_gnupg_table ();
	gettext_free_current_domain ();
        hotkeys_unregister (hwnd);
        PTD_delete ();
	agent_flush_cache ();
	html_help_deinit (cookie);
	wipe_clipboard (manual_shutdown);
	debug_end ();
        Shell_NotifyIcon (NIM_DELETE, &NID);
        PostQuitMessage (0);
        ExitProcess (0);
	break;
        
    case WM_USER:
        switch (lparam) {
        case WM_RBUTTONUP:
	case WM_LBUTTONUP:
            HMENU hm, popup;
            POINT p;

	    if (!wait_for_progress_cb (0))
		break;
	    gpg_read_only = gpg_check_permissions (0) == 0? 0 : 1;
            SetForegroundWindow (hwnd);
            GetCursorPos (&p);
            hm = LoadMenu (glob_hinst, MAKEINTRESOURCE (IDR_WINPT));
            popup = GetSubMenu (hm, 0);

	    if (!scard_support)
		set_menu_state (popup, ID_WINPT_CARD, MF_DISABLED|MF_GRAYED);
            set_menu_text (popup, ID_WINPT_FILE, _("File Manager"));
            set_menu_text (popup, ID_WINPT_KEY, _("Key Manager"));
	    set_menu_text (popup, ID_WINPT_CARD, _("Card Manager"));
            set_menu_text (popup, ID_WINPT_EDIT, _("Edit"));
            set_menu_text (popup, ID_WINPT_ABOUT, _("About..."));
            set_menu_text (popup, ID_WINPT_QUIT, _("Exit"));
            set_menu_text (popup, ID_WINPT_SYMENC, _("Symmetric"));
            set_menu_text (popup, ID_WINPT_ENCRYPT, _("Encrypt"));
	    set_menu_text (popup, ID_WINPT_SIGN, _("Sign"));
            set_menu_text (popup, ID_WINPT_SIGNENCRYPT, _("Sign && Encrypt"));
            set_menu_text (popup, ID_WINPT_DECRYPT, _("Decrypt/Verify"));
            set_menu_text (popup, ID_WINPT_VERIFY, _("Verify"));
            set_menu_text (popup, ID_WINPT_CURRWND_ENCRYPT, _("Encrypt"));
            set_menu_text (popup, ID_WINPT_CURRWND_SIGNENCRYPT, _("Sign && Encrypt"));
            set_menu_text (popup, ID_WINPT_CURRWND_DECRYPT_VERIFY, _("Decrypt/Verify"));
            set_menu_text (popup, ID_WINPT_CURRWND_SIGN, _("Sign"));
	    set_menu_text (popup, ID_WINPT_CURRWND_SYMENC, _("Symmetric"));
	    if (!secret_key_available ()) {
		set_menu_state (popup, ID_WINPT_SIGN, MF_DISABLED|MF_GRAYED);
		set_menu_state (popup, ID_WINPT_CURRWND_SIGN, MF_DISABLED|MF_GRAYED);
		set_menu_state (popup, ID_WINPT_SIGNENCRYPT, MF_DISABLED|MF_GRAYED);
		set_menu_state (popup, ID_WINPT_CURRWND_SIGNENCRYPT, MF_DISABLED|MF_GRAYED);
	    }
	    /* change popup texts */	  
	    set_menu_text_bypos (popup, 4, _("Clipboard"));
	    set_menu_text_bypos (popup, 5, _("Current Window"));
	    set_menu_text_bypos (popup, 7, _("Preferences"));

            TrackPopupMenu (popup, TPM_RIGHTALIGN, p.x, p.y, 0, hwnd, NULL);
            PostMessage (hwnd, WM_USER, 0, 0);
            DestroyMenu (popup);
            DestroyMenu (hm);
            break;
        }
        break;
        
    case WM_HOTKEY:
        cmd = 0;
        switch (wparam) {
        case WPT_CLIP_ENCRYPT_ID:			
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_ENCRYPT, 0 );
            break;
            
        case WPT_CLIP_DECRYPT_VERIFY_ID:
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_DECRYPT_VERIFY, 0 );
            break;
            
        case WPT_CLIP_SIGN_ID:
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_SIGN, 0 );
            break;
            
        case WPT_CLIP_SIGN_ENCRYPT_ID:
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_SIGNENCRYPT, 0 );
            break;
            
        case WPT_CURRWND_ENCRYPT_ID:
            use_hotkey = 1;
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_CURRWND_ENCRYPT, 0 );
            break;				
            
        case WPT_CURRWND_SIGN_ID:
            use_hotkey = 1;
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_CURRWND_SIGN, 0 );
            break;
            
        case WPT_CURRWND_SIGN_ENCRYPT_ID:
	    use_hotkey = 1;
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_CURRWND_SIGNENCRYPT, 0 );
            break;
            
        case WPT_CURRWND_DECRYPT_VERIFY_ID:
            use_hotkey = 1;
            SendMessage( hwnd, WM_COMMAND, ID_WINPT_CURRWND_DECRYPT_VERIFY, 0 );
            break;
            
        case WPT_AGENT_FORGET_ID:
            rc = msg_box (hwnd, _("Remove all passphrases from cache?"),
			  _("WinPT"), MB_YESNO|MB_ICONQUESTION);
            if (rc == IDYES)
                agent_flush_cache ();
            break;
        }		
        break;
        
    case WM_COMMAND:
        if (gnupg_access_keyring (1)) {
	    if (log_box( "WinPT", MB_ERR|MB_YESNO,
		_("Could not access public keyring, exit WinPT?") ) == IDYES)
                SendMessage (hwnd, WM_DESTROY, 0, 0);
            break;
        }
        switch (wparam) {
            /** Clipboard operations **/
        case ID_WINPT_SYMENC:
        case ID_WINPT_ENCRYPT:
        case ID_WINPT_SIGN:
        case ID_WINPT_SIGNENCRYPT:
        case ID_WINPT_DECRYPT_VERIFY:
	    err = gpg_clip_istext_avail (&has_data);
	    if (err) {
		msg_box (hwnd, gpgme_strerror (err),_("Clipboard"), MB_ERR);
		break;
	    }
	    if (!has_data) {
		msg_box (hwnd, winpt_strerror (WPTERR_CLIP_EMPTY),
			 _("Clipboard"), MB_ERR);
		break;
	    }
            clip_gpg_dlg (hwnd, (int)wparam, NULL);
            break;

            /** Current window operations **/
	case ID_WINPT_CURRWND_SYMENC:
        case ID_WINPT_CURRWND_ENCRYPT:
        case ID_WINPT_CURRWND_SIGNENCRYPT:
        case ID_WINPT_CURRWND_DECRYPT_VERIFY:
        case ID_WINPT_CURRWND_SIGN:
            rc = copy_window_content (hwnd, &cw_main, &cw_focus, use_hotkey);
            if (rc) {
		log_box (_("WinPT Error"), MB_ERR, 
			 _("Make sure that the window contains text.\n"
			   "%s."),
			 winpt_strerror (WPTERR_CURR_WND));
                break;
            }
            clip_gpg_dlg (hwnd, (int)wparam, &set_wc);
            if (set_wc == 1)
                paste_window_content (hwnd, cw_main, cw_focus);
	    use_hotkey = 0;
            break;
            
            /** File handling **/
        case ID_WINPT_FILE:
	    s = _("File Manager (use drag & drop to add files)");
	    wnd = FindWindow (NULL, s);
	    if (wnd && IsIconic (wnd))
		ShowWindow (wnd, SW_RESTORE);
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_FILE,
		              GetDesktopWindow (), file_manager_dlg_proc, 0,
                              s, IDS_WINPT_FILE);
            break;
            
            /** Misc **/
        case ID_WINPT_KEY:
	    wnd = FindWindow (NULL, _("Key Manager"));
	    if (wnd && IsIconic (wnd))
		ShowWindow (wnd, SW_RESTORE);
	    if (wnd)
		SetForegroundWindow (wnd);
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYMISC, 
			      GetDesktopWindow(), keymanager_dlg_proc, 
			      cmd != 0? (LPARAM)cmd : 0, 
                              _("Key Manager"), IDS_WINPT_KEYMISC);
            cmd = 0;
            break;

	case ID_WINPT_CARD:
	    load_card_manager ();
	    break;
            
        case ID_WINPT_EDIT:
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_CLIPEDIT,
			      GetDesktopWindow(), clip_edit_dlg_proc, 0,
                             _("Clipboard Editor"), IDS_WINPT_CLIPEDIT );
            break;
            
        case ID_WINPT_ABOUT:
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_ABOUT, GetDesktopWindow(), 
                              about_winpt_dlg_proc, 0, 
                             _("About WinPT"), IDS_WINPT_ABOUT );
            break;
            
        case ID_WINPT_PREFS:
            dialog_box_param( glob_hinst, (LPCSTR)IDD_WINPT_PREFS, GetDesktopWindow(),
                              prefs_dlg_proc, 0,
                             _("WinPT Preferences"), IDS_WINPT_PREFS );
            break;
            
        case ID_WINPT_GPGPREFS:
            rc = dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_GPGPREFS, 
				    GetDesktopWindow(), gpgprefs_dlg_proc, 0,
                              _("GnuPG Preferences"), IDS_WINPT_GPGPREFS);
	    if (rc == TRUE)
		update_keycache (GetDesktopWindow ());
            break;
            
        case ID_WINPT_QUIT:
	    manual_shutdown = 1;
            SendMessage (hwnd, WM_DESTROY, 0, 0);
            break;
        }
        break;
    }
    
    return DefWindowProc (hwnd, msg, wparam, lparam);
}
